%
% MALHEUR - Automatic Analysis of Malware Behavior
% Copyright (c) 2009-2012 Konrad Rieck (konrad@mlsec.org)
% University of Goettingen, Berlin Institute of Technology
% 
% Minimalistic interface to Malheur functionality
%
% Synopsis:
%    out = malheur(cmd, args);
%
% Arguments:
%     cmd:      Command to execute
%        
%          'mal_distance'   -  See help of mal_distance.m
%          'version'        -  Print Malheur version
%          'verbose'        -  Increase verbosity
%
